/*
 * ============================================================================
 * Name        : smokers.c
 * Author      : Alen Stojanov
 * Version     : 1.0
 * Copyright   : GNU GPL Licence
 * Description : Conditional Variables POSIX
 *============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

/*
 * We can represent the smokers problem with 4 cases:
 * 0 - there are not ingredients on the table
 * 1 - there are only tobacco and matches on the table
 * 2 - there are only tobacco and paper on the table
 * 3 - there are only paper and matches on the table.
 */

#define NONE            0
#define MATCHES_TOBACCO 1
#define PAPER_TOBACCO   2
#define MATCHES_PAPER   3

/* Number of smoker threads. */

#define NUM_SMOKER      3

static char *const status[] = {
    "NONE",
    "MATCHES_TOBACCO",
    "PAPER_TOBACCO",
    "MATCHES_PAPER"
};

typedef struct {
    int status;
    pthread_mutex_t mutex;
    pthread_cond_t  cond_empty;
    pthread_cond_t  cond_full;
} table_t;

static table_t table;

static void* agent_thread(void* args)
{
    while (1) {
        pthread_mutex_lock(&table.mutex);
        while (table.status != NONE) {
            pthread_cond_wait(&table.cond_empty, &table.mutex);
        }
        table.status = (int)((double)rand() / RAND_MAX * 3) + 1;
        printf("[A]: Current status of the table: [%d]: %s\n",
               table.status, status[table.status]);
        pthread_cond_broadcast( &table.cond_full );
        pthread_mutex_unlock( &table.mutex );
    }
    
    return NULL;
}

static void* smoker_thread(void* args)
{
    int req = *((int*)args);
    while(1) {
        pthread_mutex_lock(&table.mutex);
        while (table.status == NONE) {
            pthread_cond_wait(&table.cond_full, &table.mutex);
        }
        if (table.status == req) {
            table.status = NONE;
            printf("[S]: The smoker requiring %s is now smoking\n",
                   status[req]);
            pthread_cond_broadcast(&table.cond_empty);
        }
        pthread_mutex_unlock(&table.mutex);
        
    }
    return NULL;
}

int main(int argc, char* argv[])
{
    int i, stat;
    int smoker_requires[NUM_SMOKER];
    pthread_t smoker_tid[NUM_SMOKER];
    pthread_t agent_tid;
    
    table.status = NONE;
    pthread_mutex_init(&table.mutex, NULL);
    pthread_cond_init(&table.cond_empty, NULL);
    pthread_cond_init(&table.cond_full, NULL);
    
    stat = pthread_create(&agent_tid, NULL, agent_thread, NULL);
    if (stat != 0) {
        fprintf(stderr, "%s: pthread_create() failed: %d\n", argv[0], stat);
    }
    
    for (i = 0; i < NUM_SMOKER; i++) {
        switch (i % 3) {
        case 0:
            smoker_requires[i] = MATCHES_TOBACCO;
            break;
        case 1:
            smoker_requires[i] = PAPER_TOBACCO;
            break;
        case 2:
            smoker_requires[i] = MATCHES_PAPER;
            break;
        }
        stat = pthread_create(&smoker_tid[i], NULL,
                              smoker_thread, (void*)&smoker_requires[i]);
        if (stat != 0) {
            fprintf(stderr, "%s: pthread_create() failed: %d\n",
                    argv[0], stat);
        }
    }
    
    pthread_join(agent_tid, NULL);
    
    for (i = 0; i < NUM_SMOKER; i++) {
        pthread_join(smoker_tid[i], NULL);
    }
    
    pthread_mutex_destroy(&table.mutex);
    pthread_cond_destroy(&table.cond_empty);
    pthread_cond_destroy(&table.cond_full);
    
    return EXIT_SUCCESS;
}
